/*
 * $Id: MenuManager.java 3476 2008-04-26 22:24:58Z uckelman $
 *
 * Copyright (c) 2008 by Joel Uckelman 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License (LGPL) as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, copies are available
 * at http://www.opensource.org.
 */

package VASSAL.tools.menu;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.swing.Action;
import javax.swing.JFrame;
import javax.swing.JMenuBar;

public abstract class MenuManager {
  protected static MenuManager instance;

  public static MenuManager getInstance() {
    if (instance == null) throw new IllegalStateException();
    return instance;
  }

  public MenuManager() {
    if (instance != null) throw new IllegalStateException();
    instance = this;
  }

  public abstract JMenuBar getMenuBarFor(JFrame fc);

  public abstract MenuBarProxy getMenuBarProxyFor(JFrame fc);

  private Map<String,List<MenuItemProxy>> actionLocations =
    new HashMap<String,List<MenuItemProxy>>();

  public MenuItemProxy addKey(String key) {
    List<MenuItemProxy> items = actionLocations.get(key);
    if (items == null) {
      items = new ArrayList<MenuItemProxy>();
      actionLocations.put(key, items);
    }

    final MenuItemProxy item = new MenuItemProxy();
    items.add(item);    
    return item;
  }

  public List<MenuItemProxy> getItems(String key) {
    return actionLocations.get(key);
  }

  public void addAction(String key, Action a) {
    final List<MenuItemProxy> items = actionLocations.get(key);
    if (items != null) {
      for (MenuItemProxy i : items) {
        i.setAction(a);
      }
    }
  }

  public void removeAction(String key) {
    addAction(key, null);
  }

  public MenuMarker getMarker(String key) {
    final List<MenuItemProxy> items = actionLocations.get(key);
    if (items != null && items.size() > 0) {
      final MenuItemProxy mip = items.get(0);
      if (mip instanceof MenuMarker) return (MenuMarker) mip;
    }
    return null;     
  }

  public MenuMarker addMarker(String key) {
    List<MenuItemProxy> items = actionLocations.get(key);
    if (items == null) {
      items = new ArrayList<MenuItemProxy>();
      actionLocations.put(key, items);
    }

    final MenuMarker marker = new MenuMarker();
    items.add(marker);
    return marker;
  }

  public void addToSection(String key, ChildProxy<?> item) {
    final MenuMarker start = getMarker(key + ".start");
    final MenuMarker end = getMarker(key + ".end");
    final MenuProxy parent = (MenuProxy) end.getParent();
    final int startPos = parent.getIndex(start);
    final int endPos = parent.getIndex(end);

    // create the separator if this is the first item
    if (startPos + 1 == endPos) {
      parent.insertSeparator(endPos+1);
    }

    // insert the item between the markers
    parent.insert(item, endPos);
  }

  public void removeFromSection(String key, ChildProxy<?> item) {
    final MenuMarker start = getMarker(key + ".start");
    final MenuMarker end = getMarker(key + ".end");
    final MenuProxy parent = (MenuProxy) end.getParent();

    // remove the item
    parent.remove(item);
      
    // remove the separator if this was the last item
    final int startPos = parent.getIndex(start);     
    final int endPos = parent.getIndex(end);     
    if (startPos + 1 == endPos) {
      parent.remove(endPos+1);
    }
  }
}
